__all__ = ['acl', 'hooks']

from functools import wraps

from _term import Hict, gen_uuid_bytes
from _exc import Unauthorized
import _config as conf

acl_rules = dict(role=lambda s, roles, *args, **argd: s.role in roles and role in s.profile.roles,  # noqa
                 user=lambda s, users, *args, **argd: s.profile.username in users,                  # noqa
                 origin=lambda s, origins, *args, **argd: s.origin in origins,
                 domain=lambda s, domains, *args, **argd: s.domain in domains)


def set_acl_rule(name, func):
    acl_rules[name] = func


def acl(**rules):

    for k, v in rules.iteritems():
        validator = acl_rules.get(k, v)
        assert callable(validator), ('invalid-acl', k)
        rules[k] = (v, validator)

    def decorator(f):

        from _state import get_state

        @wraps(f)
        def wrapper(*args, **argd):
            _state_ = get_state()
            for k, (v, validator) in rules.iteritems():
                try:
                    if v is validator:
                        valid = validator(_state_, *args, **argd)
                    else:
                        valid = validator(_state_, v, *args, **argd)
                except:
                    valid = False

                if valid is True:
                    continue
                elif valid is False:
                    raise Unauthorized(k)
                else:
                    raise Unauthorized(k, valid)

            return f(*args, **argd)

        return wrapper

    return decorator


def _undefined(name):

    def wrapper(*argv, **argd):
        raise NotImplementedError(name)

    return wrapper


def _default(key):
    key = key.upper()

    assert hasattr(conf, key), key

    def _default(*args, **argd):
        return getattr(conf, key)
    _default.is_default = True

    return _default

hooks = Hict(load_state=_undefined('hooks.load_state'),
             store_state=_undefined('hooks.store_state'),
             active_user_sessions=_undefined('hooks.active_user_sessions'),
             destroy_user_sessions=_undefined('hooks.destroy_user_sessions'),
             max_active_user_sessions=_default('max_active_user_sessions'),
             fetch_profile=_undefined('hooks.fetch_profile'),
             has_permission=_undefined('hooks.has_permission'),
             store_prelogin=_undefined('hooks.store_prelogin'),
             incr_abuse=_undefined('hooks.incr_abuse'),
             max_pads_count=_default('max_pads_count'),
             pad_grace_time=_default('pad_grace_time'),
             max_origin_abuse_level=_default('max_origin_abuse_level'),
             abuse_level_watch_period=_default('abuse_level_watch_period'),
             abuse_level_by_origin=_undefined('hooks.abuse_level_by_origin'),
             prelogin_grace_time=_default('prelogin_grace_time'),
             prelogin_count=_undefined('hooks.prelogin_count'),
             max_prelogin_count=_default('max_prelogin_count'),
             session_timeout=_default('session_timeout'),
             gen_session_id=gen_uuid_bytes)

default_settings = tuple(k for k, v in hooks.items()
                         if getattr(v, 'is_default', False))

### Profiles model reference implementation.

from time import time as timestamp

import logging
logger = logging.getLogger('fanery.auth')


class Profile:

    @classmethod
    def initialize(cls, record):
        from _state import get_state
        record.merge(domain=get_state().domain)

    @classmethod
    def validate(cls, record):
        from _term import is_str, is_email, Hict

        errors, dct = Hict(), record._dct

        domain = dct.get('domain')
        if not domain:
            errors.domain.required = domain

        username = dct.get('username')
        if not username:
            errors.username.required = username
        elif not is_str(username):
            errors.username.bad_type = type(username).__name__

        password_hash = dct.get('password_hash')
        if not password_hash:
            errors.password.required = password_hash

        email = dct.get('email')
        if not email:
            errors.email.required = email
        elif not is_str(email):
            errors.email.bad_type = type(email).__name__
        elif not is_email(email):
            errors.email.bad_format = email

        if errors:
            print dct
        return errors

    @classmethod
    def index(cls, record):
        return dict((k, v) for k, v in record._dct.iteritems()
                    if not k.startswith('password_'))


def setup(db):
    from _store import add_model, add_storage

    start = timestamp()
    try:
        add_model(Profile)
        add_storage(db, Profile)
    finally:
        logger.debug('%0.6f setup' % (timestamp() - start))


def add_user(username, password=None, **extra):
    from _crypto import nacl_random
    from _state import get_state
    from _store import storage
    from _term import is_str
    from _auth import hooks
    from scrypt import hash

    start = timestamp()
    try:
        state = get_state()
        domain = state.domain = extra.setdefault('domain', state.domain)
        extra.setdefault('email', '%s@%s' % (username, 'example.com'))

        user = hooks.fetch_profile(domain, username)
        if user is None:
            if not password:
                password = nacl_random(16).encode('hex')
            else:
                assert is_str(password), 'bad-type'

            salt = nacl_random(16)
            password_hash = hash(password, salt)

            desc = "adduser %s:%s@%s" % (username, password, domain)
            with storage(desc) as db:
                db.insert(Profile,
                          username=username,
                          password_hash=password_hash,
                          password_salt=salt,
                          **extra)
    finally:
        logger.debug('%0.6f init' % (timestamp() - start))



"""
ElasticSearch powered storage implementation.
"""

from elasticsearch import Elasticsearch as ES, serializer
from elasticsearch.helpers import bulk

from time import time as timestamp
from _store import Record
from _auth import hooks
from _term import to_simple, UUID

'''
# monkey patch elasticsearch serializer
def dumps(term, *args, **argd):
    simple = to_simple(term)
    json = to_json(simple)
    return json

serializer.json.dumps = dumps
'''

Prelogin = 'Prelogin'
Session = 'Session'
Abuse = 'Abuse'


class ElasticSearch(object):

    def __init__(self, *args, **argd):
        self._es = ES(*args, **argd)

    def do_nothing(self, *args, **argd):
        pass
    begin = commit = rollback = add_model = do_nothing

    def _bulk_actions(self, index, txn, inserts=None, updates=None, deletes=None):  # noqa
        for (action, records, source) in (('create', inserts, True),
                                          ('update', updates, True),
                                          ('delete', deletes, False)):
            if records:
                for record in records:
                    if source:
                        source = to_simple(record._dct)
                        source.update(_txn=str(txn))
                    yield dict(_index=index,
                               _op_type=action,
                               _type=record._model,
                               _id=str(record._uuid),
                               _version=record._vsn,
                               _source=source or None)

    def store(self, domain, txn, inserts=None, updates=None, deletes=None):
        actions = self._bulk_actions(domain, txn, inserts, updates, deletes)
        bulk(self._es, actions, raise_on_error=True)

    def fetch(self, domain, model, uuid, vsn=None, txn=None):
        document = self._es.get(index=domain, doc_type=model, id=uuid, version=vsn) # noqa
        _source = document['_source']
        _version = document['_version']
        _txn = _source.pop('_txn', None)
        if _txn is not None:
            _txn = UUID(_txn)
        assert not vsn or _version == vsn, 'version-mismatch'
        assert not txn or _txn == txn, 'txn-mismatch'
        return _version, _txn, _source

    def query(self, domain, model, fun=None, score=None, source=False):
        es = self._es
        es.indices.create(index=domain, ignore=400)
        result = fun(domain, model, es)
        print result
        for document in result['hits']['hits']:
            _source = document['_source']
            _version = document['_version']
            _txn = _source.pop('_txn', None)
            if _txn is not None:
                _txn = UUID(_txn)
            _id = UUID(document['_id'])
            if source is True:
                yield (_id, _version, _txn, _source)
            elif score is not None:
                #TODO: extract real document score
                yield (_id, _version, _txn, score)
            else:
                yield (_id, _version, _txn)

    def search(self, domain, model, fun):
        return self.query(domain, model, fun, 1)

    def select(self, domain, model, fun=None):
        return self.query(domain, model, fun)

    def find(self, domain, model, terms):
        query = ' AND '.join('%s:"%s"' % i for i in terms.items())
        search = lambda _d, _m, es: es.search(index=domain, doc_type=model, q=query)  # noqa
        for uuid, vsn, txn, dct in self.query(domain, model, search):
            return Record(model, uuid, vsn, txn, dct)

    def fetch_profile(self, domain, identity, **extra):
        extra.update(username=identity)
        return self.find(domain, 'Profile', extra)

    def has_permission(self, state, call):
        return True

    def active_user_sessions(self, domain, identity):
        query = 'identity:"%s" AND expires:[%s TO *]' % (identity, timestamp())
        return self._es.count(index=domain, doc_type=Session, q=query)

    def destroy_sessions(self, domain, identity):
        body = {'term': {'identity': identity}}
        self._es.delete_by_query(index=domain, doc_type=Session, body=body)

    def load_state(self, domain, sid):
        return self._es.get(index=domain, doc_type=Session, id=sid, source=True)    # noqa

    def store_state(self, state):
        self._es.index(index=state.domain, doc_type=Session,
                       id=state.sid, body=state)

    def destroy_state(self, state):
        self._es.delete(index=state.domain, doc_type=Session, id=state.sid)

    def store_prelogin(self, domain, sid, **argd):
        self._es.create(index=domain, doc_type=Prelogin, id=sid,
                        body=to_simple(argd), raise_on_error=True)

    def fetch_prelogin(self, origin, domain, sid):
        es = self._es
        if es.exists(index=domain, doc_type=Prelogin, id=sid):
            prelogin = es.get(index=domain, doc_type=Prelogin, id=sid, source=True) # noqa
            es.delete(index=domain, doc_type=Prelogin, id=sid)
            return prelogin if prelogin['origin'] == origin else None

    def prelogin_count(self, domain, origin):
        query = 'origin:"%s" AND expires:[%s TO *]' % (origin, timestamp())
        return self._es.count(index=domain, doc_type=Prelogin, q=query)

    def abuse_level_by_origin(self, origin, domain):
        watch_period = timestamp() - hooks.abuse_level_watch_period(domain)
        query = 'origin:"%s" AND expires:[%s TO *]' % (origin, watch_period)
        docs = self._es.search(index=domain, doc_type=Abuse, q=query, source=True)  # noqa
        return sum(d['incr'] for d in docs)

    def incr_abuse(self, origin, domain, incr=1, abuse=None):
        body = dict(origin=origin, incr=incr, abuse=abuse, expires=timestamp())
        self._es.index(index=domain, doc_type=Abuse, body=body)



from inspect import getmodule
from sys import modules
from os import stat

from _term import is_module, is_str

import logging
logger = logging.getLogger('fanery.autoreload')

# modules files stats
_mf_stats = dict()


def deep_reload(module, loaded=None, parent=None):
    """Deep reload of module.

    Borrowed from
    http://stackoverflow.com/questions/5364050/reloading-submodules-in-ipython
    """
    if is_str(module):
        module = modules[module]

    if not is_module(module):
        module = getmodule(module)

    if parent is None:
        parent = module.__name__

    if parent == '__main__':
        return
    elif loaded is None:
        loaded = set()

    if module.__name__.startswith(parent):
        for name in dir(module):
            member = getattr(module, name)
            if is_module(member) and member not in loaded:
                deep_reload(member, loaded, parent)
            try:
                m_file = module.__file__
                m_mtime = stat(m_file).st_mtime
                if m_mtime > _mf_stats.get(m_file, 0):
                    reload(module)
                    _mf_stats[m_file] = m_mtime
                    logger.debug('reload module %s' % module)
                loaded.add(module)
            except:
                logger.error('reload module %s' % module, exc_info=True)
